#! /usr/bin/env python3

"""
Direction analysis of the St. Nicolas House Algorithm.

Soon be updated.
"""


class SnhaDir:
    def comp_xi(self):
        """
        Computes Xi correlation matrix for the Snha objects data.
        """
        self.xi = np.zeros((self.data.shape[1], self.data.shape[1]))
        for i, r1 in enumerate(self.data):
            for j, r2 in enumerate(self.data):
                self.xi[i, j] = self.xiCorr(self.data[r1], self.data[r2])
        self.xi = pd.DataFrame(self.xi)

    def get_xi(self):
        """
        Get the Xi correlation matrix from the Snha object.

        Returns:
            xi: Xi correlation matrix
        """
        return self.xi

    def xi_corr(self, x, y):
        """
        Computes the Xi correlation between two arrays.

        Args:
            x (pandas.Series): Data vector
            y (pandas.Series): Data vector

        Returns:
            xi: Xi correlation from x to y
        """
        # x, y = pd.Series(x), pd.Series(y)
        n = len(x)
        rk_x = x.rank()
        f = y.rank(method="max") / n
        g = (-y).rank(method="max") / n
        order = rk_x.sort_values().index
        f = f[order].to_numpy()
        A = np.mean(abs(f[0 : (n - 1)] - f[1:n])) * (n - 1) / (2 * n)
        C = np.mean(g * (1 - g))
        xi = 1 - (A / C)
        return xi



#! /usr/bin/env python3

"""
The St. Nicolas House algorithm for python.

The main module of the snha4py package. It can be used either in code or as a command line application. For further information and examples see the  [readme file](https://github.com/thake93/snha4py/tree/main/snha4py/readme.md).

"""

import pandas as pd
import numpy as np
from snha4py.SnhaNewGraph import SnhaNewGraph
from snha4py.SnhaPlot import SnhaPlot
from snha4py.SnhaDataGen import SnhaDataGen


class Snha:
    def __init__(self, data=None, graph=None):
        """
        Create an St. Nicolas House Object.

        Args:
            data (pandas.DataFrame): input data
            graph (numpy.array): adjacency matrix of a graph
        """
        self.data = data
        self.graph = graph
        self.graph_pred = None
        self.corr = None
        self.chains = []
        self.col_map = {}

    def chains2admat(self):
        """
        Converts association chains into an adjacency matrix.
        """
        g = np.zeros((self.data.shape[1], self.data.shape[1]))
        for ch in self.chains:
            for i in range(len(ch) - 1):
                g[ch[i], ch[i + 1]] = 1
        self.graph_pred = g

    def check_labels(self):
        """
        Checks for attributes and sets labels for the nodes.

        Returns:
            labels (list): list of labels for the nodes
        """
        if not (self.data is None):
            labels = list(self.data.columns)
        elif not (self.graph_pred is None):
            labels = range(self.graph_pred.shape[0])
        elif not (self.graph is None):
            labels = range(self.graph.shape[0])
        else:
            labels = []
        return labels

    def clean_sub_chains(self, chains):
        """
        Removes sub chains of the association chains.

        Args:
            chains (list): list of chains

        Returns:
            chains_clean (list): list of chains wihtout sub chains
        """
        chains_clean = []
        max_len = len(chains[0])
        for i in range(len(chains) - 1, -1, -1):
            f = False
            l = len(chains[i])
            for j in range(len(chains)):
                cur_check = chains[j]
                if l == max_len:
                    f = False
                    break
                for k in range(len(cur_check) - l + 1):
                    if chains[i] == cur_check[k : k + l] and not (len(cur_check) == l):
                        f = True
                        break
                if f:
                    break
            if not (f):
                chains_clean.append(chains[i])
        return chains_clean

    def comp_corr(self, df=None, method="pearson", in_place=True):
        """
        Computes the correlation matrix for the Snha objects data.

        Args:
            df (pandas.DataFrame): compute the correlation based on the data
            method (str):
                "pearson" (default): standard correlation coefficient
                "kendall": Kendall Tau correlation coefficient
                "spearman": Spearman rank correlation
            in_place (boolean):
                True (default): assigns self.corr the computed correlation
                False: returns the computed correlation instead

        Returns:
            Only if in_place=False: the computed correlation

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        s.create_corr_data(n=200, steps=25)
        s.comp_corr()

        cor = s.get_corr()
        print(cor)
        ```
        """
        if df is None:
            df = self.data
        if in_place:
            self.corr = df.corr(method=method)
        else:
            return df.corr(method=method)

    def conf_mat(self):
        """
        Computes the confusion matrix of the predicted and true edges of the graph.

        Returns:
            cmat (numpy.ndarray; shape: (2,2)): confusion matrix

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        s.create_corr_data(n=200, steps=25)
        s.comp_corr()
        s.st_nich_alg()

        cm = s.conf_mat()
        print(cm)
        ```
        """
        true = self.graph + self.graph.T

        pred = self.graph_pred + self.graph_pred.T
        cmat = np.zeros((2, 2))
        for i in range(true.shape[0]):
            for j in range(true.shape[1]):
                if true[i, j] != 0 and pred[i, j] != 0:
                    # TP
                    cmat[0, 0] += 1
                elif true[i, j] == 0 and pred[i, j] != 0:
                    # FP
                    cmat[1, 0] += 1
                elif true[i, j] != 0 and pred[i, j] == 0:
                    # FN
                    cmat[1, 1] += 1
                else:
                    # TN
                    cmat[0, 1] += 1

        return cmat / 2

    def create_corr_data(self, n=100, steps=15, mean=100, sd=2, noise=1, prop=0.05):
        """
        Create correlation data for the Snha objects graph.

        Args:
            n (int): the number of measurements per node, default: 100
            steps (int): the number of iterations, default: 15
            mean (float): mean value for sampling from a Gaussian distribution, default: 100
            sd (float): standard deviation for sampling from a Gaussian distribution, default: 2
            noise (float): standard deviation for sampling noise from a Gaussian distribution with mean 0 after each iteration, default: 1
            prop (float): proportion of the target node value take the source node, default: 0.05

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        s.create_corr_data(n=200, steps=25)
        data = s.get_data()
        print(data)
        ```
        """
        sdg = SnhaDataGen(
            graph=self.graph, n=n, steps=steps, mean=mean, sd=sd, noise=noise, prop=prop
        )
        self.data = sdg.get_data()

    def get_chains(self, rename=True):
        """
        Get the association chains extrected by the St. Nicholas Algorithm.

        Args:
            rename (boolean): default: True
                True: Chains of column names of the input data
                False: Chains of column index numbers

        Returns:
            chains (list): Association chains for the correlation data
        """
        if rename:
            col_map_rev = {y: x for x, y in self.col_map.items()}
            new_names = [[col_map_rev[i] for i in l] for l in self.chains]
            return new_names
        else:
            return self.chains

    def get_corr(self):
        """
        Get the correlation data from the Snha object.

        Returns:
            corr (pandas.DataFrame): Correlation matrix of the Snha obejct
        """
        return self.corr

    def get_data(self):
        """
        Get the data from the Snha object.

        Returns:
            data (pandas.DataFrame): DataFrame with m observations for n variables
        """
        return self.data

    def get_graph(self):
        """
        Get the graph from the Snha object.

        Returns:
            graph (numpy.ndarray): adjacency matrix from the Snha object
        """
        return self.graph

    def get_graph_pred(self):
        """
        Get the graph predicten from the Snha object.

        Returns:
            graph_pred (numpy.ndarray): adjacency matrix from the predicted graph
        """
        return self.graph_pred

    def graph_stats(self):
        """
        Computes the accuracy, sensitivity, specificity, BCR (Balanced Classification Rate) and MCC (Matthews Correlation Coefficient).

        Returns:
            stats (dictionary): dictionary containing the statistics

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        s.create_corr_data(n=200, steps=25)
        s.comp_corr()
        s.st_nich_alg()

        stats = s.graph_stats()
        print(stats)
        ```
        """
        cmat = self.conf_mat()
        stats = {}
        TP, TN, FP, FN = cmat[0, 0], cmat[0, 1], cmat[1, 0], cmat[1, 1]

        stats["Accuracy"] = (TP + TN) / (TP + TN + FP + FN)
        stats["Sensitivity"] = TP / (TP + FN)
        stats["Specificity"] = TN / (TN + FP)
        stats["BCR"] = (stats["Sensitivity"] + stats["Specificity"]) / 2
        stats["MCC"] = (TP * TN - FP * FN) / (
            np.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN))
        )
        return stats

    def middle_chains(self, nodes, df):
        """
        Extracts the middle chains of resulting from the correlation data.

        Args:
            nodes (list): list of nodes
            df (pandas.DataFrame): matrix of squared correlation coeficients

        Returns:
            fchain (list): middle chains of a list of nodes
        """
        nodes.reverse()
        for fi in range(1, len(nodes) - 1):
            fl = nodes[fi]
            for si in range(fi + 1, len(nodes)):
                sl = nodes[si]
                df_sort = (
                    df[[fl, sl]].iloc[nodes].sort_values(by=sl, key=lambda x: np.abs(x))
                )
                schain = df_sort.index.to_list()
                df_sort = (
                    df[[fl, sl]].iloc[nodes].sort_values(by=fl, key=lambda x: np.abs(x))
                )
                fchain = df_sort.index.to_list()
                fchain.reverse()
                if fchain == schain:
                    nodes.reverse()
                    return fchain
        nodes.reverse()

    def new_graph(self, graph_type="werner", nodes=5, edges=8, mode="directed", cont=2):
        """
        Creates a new graph for the Snha object.

        Args:
            graph_type (str): type of the graph
                "werner" (default)
                "rnd_chain"
                "band"
                "circle"
                "hub"
                "random"
                "barabasi_m1"
                "barabasi_m2"
            nodes (int): number of nodes in the graph (default: 5)
            edges (int): number of edges in the graph (default: 6)
            mode (str):
                "directed": directed graph (default)
                "undirected": undirected graph
            cont (int): number of signal seeds; default: 2 (only for graph type rndChain)

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        graph = s.get_graph()
        print(graph)
        ```
        """
        g = SnhaNewGraph(
            graph_type=graph_type, nodes=nodes, edges=edges, mode=mode, cont=cont
        )
        self.graph = g.get_graph()

    def plot_corr(self, labels=None, ax=None):
        """
        Plots the correlation matrix of the Snha data.

        Args:
            labels (list): labels for the nodes of the graph
            ax (matplotlib.axes): axes to plot the graph on

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        s.create_corr_data(n=200, steps=25)
        s.comp_corr()

        s.plot_corr()
        ```
        """
        if labels is None:
            labels = self.check_labels()

        p = SnhaPlot(adj_mat=None, corr=self.corr, labels_n=labels, ax=ax)
        p.corr()

    def plot_graph(
        self,
        layout=None,
        mode="directed",
        col="tab:blue",
        labels=None,
        labels_e=None,
        pred=True,
        ax=None,
        vs=0.15,
    ):
        """
        Plots the graph of the Snha Object.

        Args:
            labels (list): labels for the nodes of the graph
            target (matplotlib.axes): axes to plot the graph on
            pred (boolean):
                True (default): Plot the determined association chains
                False: Plot the edges from the adjacency matrix
            vs (float): vertrex size; default: 25

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        s.create_corr_data(n=200, steps=25)
        s.comp_corr()
        s.st_nich_alg()

        stats = s.graph_stats()
        print(stats)
        ```
        """
        if labels is None:
            labels = self.check_labels()

        if pred:
            p = SnhaPlot(
                adj_mat=self.graph_pred, corr=self.corr, labels_n=labels, ax=ax
            )
            mode = "undir"
        else:
            p = SnhaPlot(
                adj_mat=self.graph_pred, corr=self.corr, labels_n=labels, ax=ax
            )

        p.graph(layout=layout, mode=mode, col=col, labels_e=labels_e, vs=vs)

    def set_corr(self, c_data):
        """
        Set the correlation data for the Snha object.

        Args:
            c_data (pandas.DataFrame): Correlation matrix
        """
        self.corr = c_data

    def set_data(self, data):
        """
        Set the data for the Snha object.

        Args:
            data (pandas.DataFrame): DataFrame with m observations for n variables
        """
        self.data = data

    def set_graph(self, graph):
        """
        Set the graph for the Snha object.

        Args:
            graph (numpy.ndarray): adjacency matrix
        """
        self.graph = graph

    def set_graph_pred(self, graph):
        """
        Set the graph prediction for the Snha object.

        Args:
            graph (numpy.ndarray): adjacency matrix
        """
        self.graph_pred = graph

    def snha(self, df, alpha):
        """
        Extracts the association chains resulting from the correlation data.

        Args:
            df (pandas.DataFrame): correlation matrix
            alpha (float, [0,1]): correlation coefficient cut off
        """
        chains = []

        # df = self.corr**2
        self.col_map = {col: i for i, col in enumerate(df.columns)}
        df = df.rename(index=self.col_map, columns=self.col_map)
        for idx, col in df.items():
            col_sort = col.sort_values(key=lambda x: np.abs(x), ascending=True)
            if len(df) > 10:
                pos_ends = col_sort.index.to_list()[-11:]
            else:
                pos_ends = col_sort.index.to_list()
            while len(pos_ends) >= 3:
                i = pos_ends[0]
                if i == idx:
                    break
                temp_df = df[[idx, i]].iloc[pos_ends]
                s = temp_df.sort_values(by=i, key=lambda x: np.abs(x))
                nodes_sorted = s.index.to_list()
                if s[i].iloc[0] < alpha:
                    pos_ends.pop(0)
                    continue
                if pos_ends == nodes_sorted[::-1]:
                    if not (nodes_sorted in chains) and not (
                        nodes_sorted[::-1] in chains
                    ):
                        chains.append(nodes_sorted)
                        break
                else:
                    mc = self.middle_chains(pos_ends, df)
                    if mc and (not (mc in chains) and not (mc[::-1] in chains)):
                        chains.append(mc)
                        break
                pos_ends.pop(0)
        if len(chains) > 0:
            self.chains = self.clean_sub_chains(chains)
        self.chains2admat()

    def snha_bt(self, df, alpha, n, lbd, method):
        """
        Applying the St Nicholas algorithm within a bootstrapp routine.

        Args:
            df (pandas.DataFrame): Input data
            alpha (float, [0,1]): correlation coefficient cut off
            n (int): number of bootstrap iterations
            lbd (float, [0,1]): fraction of all iterations to accept an edge as a prediction (e.g. 50/100 iterations an edge need to be found, to be considered as a predicted edge at lbd=0.5 and n=100)
            method (str):
                "pearson" (default): standard correlation coefficient
                "kendall": Kendall Tau correlation coefficient
                "spearman": Spearman rank correlation
        """
        predictions = []
        samples = [df.sample(len(df), replace=True) for i in range(n)]
        for i, smpl in enumerate(samples):
            c = self.comp_corr(df=smpl, in_place=False, method=method)
            self.snha(c**2, alpha)
            predictions.append(self.graph_pred)

        edges = []
        for p in predictions:
            a = np.where(p == 1)
            edges += list(zip(a[0], a[1]))

        edge_count = {}
        for e in edges:
            if e in edge_count or e[::-1] in edge_count:
                try:
                    edge_count[e] += 1
                except:
                    edge_count[e[::-1]] += 1
            else:
                edge_count[e] = 1

        self.comp_corr(method=method)
        self.snha(self.corr**2, alpha)
        for k, v in edge_count.items():
            if v >= 2 * n * lbd:
                x, y = k
                self.graph_pred[x, y] = 1
                self.graph_pred[y, x] = 1

    def st_nich_alg(
        self, data=None, alpha=0.1, bt=False, n=100, lbd=0.5, method="pearson"
    ):
        """
        Selection to use the St. Nicholaus algorithm with or without bootstrapping.

        Args:
            data (pandas.DataFrame):
                bt=True: Input data
                bt=False: Correlation matrix
            alpha (float, [0,1]): correlation coefficient cut off; default: 0.1
            bt (boolean): Bootstrap True/Flase; default: False
            n (int): number of bootstrap iterations; default: 100
            lbd (float, [0,1]): fraction of all iterations to accept an edge as a prediction; default: 0.5 (e.g. 50/100 iterations an edge need to be found, to be considered as a predicted edge at lbd=0.5 and n=100)
            method (string):
                "pearson" (default): standard correlation coefficient
                "kendall": Kendall Tau correlation coefficient
                "spearman": Spearman rank correlation

        Examples:

        ```{.py}
        from snha4py.Snha import Snha

        s = Snha()
        s.new_graph()
        s.create_corr_data(n=200, steps=25)
        s.comp_corr()
        s.st_nich_alg()

        chains = s.get_chains()
        graph_pred = s.get_graph_pred()

        print(chains)
        print(graph_pred)
        ```
        """
        if bt:
            if data is None:
                data = self.data
            self.snha_bt(data, alpha**2, n, lbd, method)
        else:
            if data is None:
                data = self.corr
            self.snha(data**2, alpha**2)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    s = Snha()
    s.new_graph()
    s.create_corr_data(
        n=200,
        steps=25,
        mean=100,
        sd=2,
        noise=1,
        prop=0.05,
    )
    s.comp_corr()
    s.st_nich_alg()
    s.data.columns = ["A", "B", "C", "D", "E", "F"]
    s.plot_graph(pred=False)
    s.plot_graph()
    s.plot_corr()
    plt.show()



#! /usr/bin/env python3

"""
API of snha4py

Documentation of the St. Nicolas House algorithm for python3. 

Classes

```{.kroki echo=false dia=plantuml}
class Snha{
+ data
+ graph
+ graph_pred
+ corr
+ chains
- col_map

- chains2admat()
- check_labels()
- clean_sub_chains(chains)
+ comp_corr(df, mehtod, in_place)
+ conf_mat()
+ create_corr_data(n, steps, mean, sd, noise, prop)
+ get/set() for all attributes
+ graph_stats()
- middle_chains(nodes, df)
+ new_graph(graph_type, nodes, edges, mode, cont)
+ plot_corr(labels, ax)
+ plot_graph(layout, mode, col, labels, labels_e, pred, ax)
- snha(df, alpha)
- snha_bt(df, alpha, n, lbd, method)
+ st_nich_alg(data, alpaha, bt, n, lbd, method)
}

class SnhaDataGen{
+ graph
+ n
+ steps
+ mean
+ sd
+ noise
+ prop
 
+ create_data()
+ get_data()
}

class SnhaNewGraph{
+ nodes
+ edges
- graph
+ mode
+ cont

- band()
- barabasi(m)
- barabasi_m1()
- barabasi_m2()
- circle()
- edge_dir_shuffle()
- hub()
+ get_graph()
- random()
- rnd_chain()
+ set_graph(fct)
- undir2dir(start, visited, g_new, nodes)
- werner()
}

class SnhaPlot{
+ data
+ labels_n
+ ax

+ corr()
- get_orth_vec(vec, scale)
+ graph(layout, mode, col, labels_e)
- scale(layout, a, b)
- unit_length_vec
}


Snha <-- SnhaDataGen
Snha <-- SnhaPlot
SnhaNewGraph -> Snha
```
"""



#! /usr/bin/env python3

"""
The St. Nicolas House Algorithm graph collection.

Examples:

```{.py}
from snha4py.SnhaNewGraph import SnhaNewGraph

fct = dir(SnhaNewGraph)
ignore = ["set_graph", "get_graph", "undir2dir"]
for f in fct:
    if f not in ignore and "__" not in f:
        print(f)
        g = SnhaNewGraph(graph_type=f, nodes=5, edges=8, mode="directed", cont=2)
        graph = g.get_graph()
        print(graph)
```
"""
import numpy as np


class SnhaNewGraph:
    def __init__(self, graph_type, nodes, edges, mode, cont):
        """
        Creates a new graph.

        Args:
            graph_type (str): type of the graph
                "werner" (default)
                "rnd_chain"
                "band"
                "circle"
                "hub"
                "random"
                "barabasi_m1"
                "barabasi_m2"
            nodes (int): number of nodes in the graph (default: 5)
            edges (int): number of edges in the graph (default: 6)
            mode (str):
                "directed": directed graph (default)
                "undirected": undirected graph
            cont (int): number of signal seeds; default: 2 (only for graph type rndChain)
        """
        self.nodes = nodes
        self.edges = edges
        self.graph = np.zeros((nodes, nodes))
        self.mode = mode
        self.cont = cont
        self.set_graph(graph_type)

    def band(self):
        """
        Implementation of a band graph.
        """
        for i in range(self.nodes - 1):
            self.graph[i, i + 1] = 1

    def barabasi(self, m=1):
        """
        Creates an Barabasi-M1/M2 graph.

        Args:
            m (int):
                1 (default): barabasi-M1 graph
                2: brabasi-M2 graph
        """
        self.graph[1, 0] = 1
        for n in range(2, self.nodes):
            if m == n:
                sel = n - 1
            else:
                sel = m
            deg = (self.graph + self.graph.T).sum(axis=1)
            prob = deg[:n] / (deg[:n].sum())
            idx = np.random.choice(range(n), size=sel, p=prob)
            self.graph[idx, n] = 1

    def barabasi_m1(self):
        """
        Implementation of a barabasi-M1 graph.
        """
        self.barabasi()

    def barabasi_m2(self):
        """
        Implementation of a barabasi-M2 graph.
        """
        self.barabasi(m=2)

    def circle(self):
        """
        Implementation of a circle graph.
        """
        self.band()
        self.graph[self.nodes - 1, 0] = 1

    def edge_dir_shuffle(self):
        """
        Shuffle the direction of the Snha objects graph edges.
        """
        for i in range(self.graph.shape[0]):
            for j in range(self.graph.shape[0]):
                w = self.graph[i, j] + self.graph[j, i]
                if w > 0:
                    if int(np.random.uniform(0, 2)):
                        self.graph[i, j] = w
                        self.graph[j, i] = 0
                    else:
                        self.graph[i, j] = 0
                        self.graph[j, i] = w

    def hub(self):
        """
        Implementation of a hub graph.
        """
        ix = np.random.choice(range(self.nodes))
        self.graph[ix, :] = 1
        self.graph[ix, ix] = 0

    def get_graph(self):
        """
        Returns:
            self.graph (numpy.ndarray): the created graph
        """
        if self.mode == "undirected":
            return self.graph + self.graph.T
        else:
            return self.graph

    def random(self):
        """
        Implementation of a random graph.
        """
        g = np.zeros(self.nodes**2)
        ix = np.random.choice(range(self.nodes**2), self.edges, replace=False)
        g[ix] = 1
        self.graph = g.reshape((self.nodes, self.nodes))

    def rnd_chain(self):
        """
        Implementation of a random graph. Select random start nodes and create directed paths through all nodes.
        """
        x, y = np.triu_indices(self.nodes, k=1)
        up_tri = np.array(list(zip(x, y)))
        v = np.random.choice(range(len(up_tri)), size=self.edges, replace=False)
        idx = up_tri[v]
        for i in idx:
            x, y = i
            self.graph[x, y] = 1
        self.graph = self.graph + self.graph.T

        start = np.random.choice(range(self.nodes), size=self.cont, replace=False)
        g_new = np.zeros((self.nodes, self.nodes))
        self.undir2dir(start, [], g_new, self.nodes)
        self.graph = g_new

    def set_graph(self, fct):
        """
        Calls the graph-function and sets the graph.

        Args:
            fct (string): name of the graph to call the function
        """
        f = getattr(self, fct)
        f()

    def undir2dir(self, start, visited, g_new, nodes):
        """
        Creates a directed graph from an undirected graph.

        Args:
            start (numpy.ndarray): list of start nodes
            visited (list): list of nodes already visited
            g_new (np.array): directed graph
            nodes (int): number of nodes
        """
        all_n = []
        np.random.shuffle(start)
        for s in start:
            if s in visited:
                continue
            visited.append(s)
            (neigh,) = np.where(self.graph[s] == 1)
            all_n.append(neigh)
            np.random.shuffle(neigh)

            for n in neigh:
                if g_new[s, n] == 1 or g_new[n, s] == 1:
                    continue
                else:
                    g_new[s, n] = 1
            if len(visited) == nodes:
                return
        for n in all_n:
            self.undir2dir(n, visited, g_new, nodes)

    def werner(self):
        """
        Implementation of a werner graph.
        """
        self.graph = np.array(
            [
                [0, 0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0, 0],
                [0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 1, 1],
                [0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0],
            ]
        )


if __name__ == "__main__":
    fct = dir(SnhaNewGraph)
    ignore = ["set_graph", "get_graph", "undir2dir"]
    for f in fct:
        if f not in ignore and "__" not in f:
            print(f)
            g = SnhaNewGraph(graph_type=f, nodes=5, edges=8, mode="directed", cont=2)
            graph = g.get_graph()
            print(graph)



import setuptools
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setuptools.setup(
    name="snha4py",  # Replace with your own username
    version="0.0.3",
    author="Tim Hake",
    author_email="timhake@icloud.com",
    description="St. Nicholas House algorithm for Python",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/thake93/snha4py",
    project_urls={
        "Documentation": "https://htmlpreview.github.io/?https://github.com/thake93/snha4py/blob/main/docs/__init__.html"
    },
    packages=["snha4py"],
    license="MIT License",
    license_files=["LICENSE"],
    # classifiers=[
    #    "Programming Language :: Python :: 3",
    #    "License :: OSI Approved :: MIT License",
    #    "Operating System :: OS Independent",
    # ],
    install_requires=[
        "numpy>=1.21.3",
        "igraph>=0.9.11",
        "pandas>=1.3.4",
        "matplotlib>=3.4.3",
    ],
)



#! /usr/bin/env python3

"""
St. Nicolas House Algorithm data generation.

The module generates data correlated data along a directed graph based on the Monte Carlo approach. For further inforation see [Novine et al. (2022)](https://doi.org/10.52905/hbph2021.3.26).
"""
import numpy as np
import pandas as pd


class SnhaDataGen:
    def __init__(self, graph, n, steps, mean, sd, noise, prop):
        """
        Create a data generator for correlated data based on a graph.

        Args:
            graph (numpy.ndarray): adjacents matrix of a directed graph
            n (int): the number of measurements per node, default: 100
            steps (int): the number of iterations, default: 15
            mean (float): mean value for sampling from a Gaussian distribution, default: 100
            sd (float): standard deviation for sampling from a Gaussian distribution, default: 2
            noise (float): standard deviation for sampling noise from a Gaussian distribution with mean 0 after each iteration, default: 1
            prop (float): proportion of the target node value take the source node, default: 0.05
        """
        self.graph = graph
        self.n = n
        self.steps = steps
        self.mean = mean
        self.sd = sd
        self.noise = noise
        self.prop = prop

    def create_data(self):
        """
        Creates correlated data along edges of a directed graph based on a Monte Carlo approach.

        Examples:

        ```{.py}
        from snha4py.SnhaDataGen import SnhaDataGen
        import numpy as np

        graph = np.array(
            [
                [0, 0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0, 0],
                [0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 1, 1],
                [0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0],
            ]
        )
        dg = SnhaDataGen(graph, n=100, steps=15, mean=100, sd=2, noise=1, prop=0.05)

        data = dg.get_data()
        print(data)
        ```
        """
        p = self.graph.shape[0]
        vtx = np.arange(p)
        data_mat = np.zeros((p, self.n))
        for i in range(self.n):
            x = np.random.normal(self.mean, self.sd, size=p)
            for j in range(self.steps):
                np.random.shuffle(vtx)
                for source in vtx:
                    neighbors = np.flatnonzero(self.graph[source])
                    if len(neighbors) == 0:
                        continue
                    np.random.shuffle(neighbors)
                    for target in neighbors:
                        weight = self.graph[source, target]
                        x_ = x[source] * self.prop * abs(weight) + x[target] * (
                            1 - self.prop
                        ) * abs(weight)
                        if weight < 0:
                            x[target] = 2 * x[target] - x_
                        else:
                            x[target] = x_
                x += np.random.normal(0, self.noise, size=p)
            data_mat[:, i] = x
        self.data = pd.DataFrame(data_mat.T)

    def get_data(self):
        """
        Returns:
            self.data (pandas.DataFrame): correlated data of size (p, n), with p the number of nodes of the graph and n the number of measurements per node.
        """
        self.create_data()
        return self.data


if __name__ == "__main__":
    graph = np.array(
        [
            [0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1, 1],
            [0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0],
        ]
    )
    dg = SnhaDataGen(graph, n=100, steps=15, mean=100, sd=2, noise=1, prop=0.05)

    data = dg.get_data()
    print(data)



#! /usr/bin/env python3

"""
The St. Nicolas House Algorithm plotting routine.
"""

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.colors import LinearSegmentedColormap
import random
import numpy as np
import igraph as ig


class SnhaPlot:
    def __init__(self, adj_mat, corr, labels_n, ax=None):
        """
        Create a support object to plot a graph or a correlation matrix.

        Args:
            adj_mat (numpy.ndarray): adjacency matrix to plot a graph with SnhaPlot.graph()
            corr (pd.DataFrame): correlation matrix to plot with SnhaPlot.corr()
            ax (matplotlib.axes): target axes to place the plot
            labels_n (list): variable names
        """
        self.adj_mat = adj_mat
        self.corr_mat = corr
        self.labels_n = labels_n

        if ax is None:
            plt.figure(figsize=(10, 10))
            self.ax = plt.gca()
        else:
            self.ax = ax

        self.ax.tick_params(left=False, bottom=False)
        self.ax.spines["top"].set_visible(False)
        self.ax.spines["right"].set_visible(False)
        self.ax.spines["bottom"].set_visible(False)
        self.ax.spines["left"].set_visible(False)

    def corr(self):
        """
        Plot the correlation matrix.
        """
        cvals = [-1, 0, 1]
        colors = ["tab:orange", "white", "tab:blue"]
        norm = plt.Normalize(min(cvals), max(cvals))
        tuples = list(zip(map(norm, cvals), colors))
        cmap = LinearSegmentedColormap.from_list("", tuples)
        mask = np.zeros_like(self.corr_mat)
        mask[np.triu_indices_from(mask, k=1)] = True
        mask = self.corr_mat * mask
        im = self.ax.imshow(mask, cmap=cmap, vmin=-1, vmax=1)
        for i in range(self.corr_mat.shape[0]):
            for j in range(self.corr_mat.shape[1]):
                if mask[j][i] != 0:
                    self.ax.annotate(
                        np.round(mask[j][i], 3),
                        xy=(i, j),
                        ha="center",
                        va="center",
                    )
        self.ax.set_title("Correlations")
        self.ax.set_xticks([i for i in range(mask.shape[0])])
        self.ax.set_yticks([i for i in range(mask.shape[1])])
        self.ax.set_xticklabels(self.labels_n)
        self.ax.set_yticklabels(self.labels_n)

    def get_orth_vec(self, vec, scale=1):
        """
        Computes an orthogonal vector of an input vector.
        Args:
            vec (numpy.ndarry): input vector
            scale (float): scales the orthogonal vector

        Returns:
            orth_vec (numpy.ndarray): scaled orthogonal vector
        """
        orth_vec = np.array([-vec[:, 1], vec[:, 0]]).T
        orth_vec = self.unit_length_vec(orth_vec)
        return orth_vec * scale

    def graph(self, layout, mode, col, labels_e, vs):
        """
        Plots a graph.
        Args:
            layout (list): list of coordinates of the nodes
            mode (string): 'directed' or 'undirected'
            col (matplotlib.colors): color of the nodes or a list of colors, which holds a color for each node
            labels_e (list): list of edge labels
            vs (float): size for the nodes
        """
        circle_radius = vs
        scale_a = -2
        scale_b = 2

        if layout is None:
            random.seed(123)
            G = ig.Graph.Adjacency(self.adj_mat)
            layout = np.array(G.layout("fr"))
        self.scale(layout, scale_a, scale_b)

        if mode == "directed":
            style = "Simple, tail_width=0.5, head_width=4, head_length=8"
        else:
            style = "Simple, tail_width=0.5, head_width=0"
            self.adj_mat = np.triu(self.adj_mat + self.adj_mat.T)

        kw = dict(arrowstyle=style, color="k")

        start, end = np.where(self.adj_mat != 0)
        # start = layout[start]
        # end = layout[end]

        edges = []
        for i in range(len(start)):
            s = start[i]
            e = end[i]
            if self.corr_mat.iloc[s, e] >= 0:
                c = "k"
            else:
                c = "red"
            edges.append(
                patches.FancyArrowPatch(
                    posA=(layout[s][0], layout[s][1]),
                    posB=(layout[e][0], layout[e][1]),
                    shrinkA=circle_radius * 200,  # 20,
                    shrinkB=circle_radius * 200,  # 20,
                    arrowstyle=style,
                    color=c,
                    # **kw
                )
            )
        for e in edges:
            self.ax.add_patch(e)

        for i, l in enumerate(layout):
            if type(col) == list:
                c = col[i]
            else:
                c = col

            n = patches.Circle(
                (l[0], l[1]),
                radius=circle_radius,
                color=c,
            )
            self.ax.add_patch(n)

        # nodes = [
        #     patches.Circle(
        #         (l[0], l[1]),
        #         radius=circle_radius,
        #         color=col,
        #     )
        #     for l in layout
        # ]
        # for n in nodes:
        #     self.ax.add_patch(n)

        for i in range(len(layout)):
            x, y = layout[i]
            self.ax.annotate(
                self.labels_n[i],
                xy=(x, y),
                ha="center",
                va="center",
                fontsize=15,
                fontweight="bold",
                color="lightgray",
            )

        if not (labels_e is None):
            true_edge = end - start
            perp_edge = self.get_orth_vec(true_edge, 0.1)
            edge_center_coords = start + true_edge * 0.5 + perp_edge
            for i in range(len(edge_center_coords)):
                x, y = edge_center_coords[i]
                self.ax.annotate(
                    labels_e[i],
                    xy=(x, y),
                    ha="center",
                    va="center",
                )

        self.ax.set_xticklabels([])
        self.ax.set_yticklabels([])
        self.ax.set_xlim(scale_a - circle_radius - 0.05, scale_b + circle_radius + 0.05)
        self.ax.set_ylim(scale_a - circle_radius - 0.05, scale_b + circle_radius + 0.05)
        self.ax.set_aspect("equal")
        plt.tight_layout()

    def scale(self, layout, a, b):
        """
        Centers and rescale the layout coordinates for the nodes.

        Args:
            layout (numpy.ndarray): layout of the nodes
            a (float): lower bound for rescaling
            b (float): upper bound for rescaling
        """
        mean = layout.mean(axis=0)
        std = layout.std()
        for i in range(len(layout)):
            layout[i] = (layout[i] - mean) / std
        mi = layout.min()
        ma = layout.max()
        for i in range(len(layout)):
            for j in range(2):
                layout[i, j] = a + ((layout[i, j] - mi) * (b - a)) / (ma - mi)

    def unit_length_vec(self, vec):
        """
        Rescales a vector to unit length.
        Args:
            vec (numpy.ndarry): Input vector

        Returns:
            new_vec (numpy.ndarry): Unit length version of the input vector
        """
        new_vec = []
        for v in vec:
            mag = np.sqrt((v**2).sum())
            new_vec.append((v[0] / mag, v[1] / mag))
        return np.array(new_vec)


if __name__ == "__main__":
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt

    data = pd.read_csv("examples/decathlon.tab", sep="\t")
    dist = [100, 110, 400, 1500]

    for d in dist:
        data[str(d)] = d / data[str(d)] * 3.6
        data = data.rename(columns={"poid": "shot", "haut": "high", "perc": "pole"})

    graph = np.array(
        [
            [0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0],
            [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
        ]
    )
    labels = data.columns
    c = data.corr("spearman")
    print(c)
    p = SnhaPlot(adj_mat=graph, corr=c, labels_n=labels)

    # c = ["blue", "red", "green", "orange", "gray", "tab:blue"]

    p.graph(None, "undirected", "tab:blue", None, 0.15)
    plt.show()



#! /usr/bin/env python3

"""
The St. Nicolas House Algorithm comando line application.

Currently available functions are:

- the general algorithm with/ without bootstrap to predict a graph from data
- create correlated data based on a directed graph
- crete a directed graph

For further information and examples see [readme file](https://github.com/thake93/snha4py/tree/main/snha4py/readme.md).
"""

from snha4py.Snha import Snha
import argparse
import numpy as np
import pandas as pd
import os
import matplotlib.pyplot as plt


def create_data(args):
    """
    A wrapper to create data based on the arguments given by the parser.

    Argumenst:
        args (argparse.Namespace): Namespace object result from the comando line arguments
    """
    f = read_file(args.graph, True)
    if f is None:
        print(
            "Use a *.tab or *.csv file containing a adjacency matrix of a directed graph."
        )
        return
    f = f.to_numpy()
    s = Snha(graph=f)
    s.create_corr_data(
        n=args.iterations,
        steps=args.steps,
        mean=args.mean,
        sd=args.std,
        noise=args.noise,
        prop=args.prop,
    )
    data = s.get_data()
    if args.output is None:
        data.to_csv("temp.csv", sep=",", index=False)
        print(f"Correlated data saved in temp.csv")
    else:
        data.to_csv(args.output, sep=",", index=False)
        print(f"Correlated data saved in {args.output}")

    if args.plot:
        s.comp_corr()

        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        s.plot_graph(pred=False, ax=ax[0])
        s.plot_corr(ax=ax[1])
        plt.show()


def create_graph(args):
    """
    A wrapper to create a graph based on the arguments given by the parser.

    Argumenst:
        args (argparse.Namespace): Namespace object result from the comando line arguments
    """
    print("graph")
    s = Snha()
    s.new_graph(
        graph_type=args.graph_type,
        nodes=args.nodes,
        edges=args.edges,
        mode=args.mode,
        cont=args.cont,
    )
    g = s.get_graph()
    if args.output is None:
        np.savetxt("temp.csv", g, delimiter=",")
        print(f"Graph was saved in temp.csv")
    else:
        np.savetxt(f"{args.output}", g, delimiter=",", fmt="%.3f")
        print(f"Graph was saved to {args.output}.")

    if args.plot:
        s.plot_graph(pred=False)
        plt.show()


def read_file(f, graph):
    """
    Reads a file.

    Arguments:
        f (str): path to the file
        graph (boolean): True: reads the file without a header; False: reads the file with a header
    Returns:
        in_file (pandas.DataFrame): data or adjacency matrix
    """
    name, ext = os.path.splitext(f)
    if ext == ".csv":
        delimiter = ","
    elif ext == ".tab":
        delimiter = "\t"
    else:
        return
    if graph:
        in_file = pd.read_csv(f, delimiter=delimiter, header=None)
    else:
        in_file = pd.read_csv(f, delimiter=delimiter)
    return in_file


def snha(args):
    """
    A wrapper to run the St. Nicolas House algorithm based on the arguments given by the parser.

    Argumenst:
        args (argparse.Namespace): Namespace object result from the comando line arguments
    """
    f = read_file(args.data, False)
    if f is None:
        print("Use a *.tab or *.csv file containing a data to analyse.")
        return

    s = Snha(data=f)
    s.comp_corr(method=args.method)
    s.st_nich_alg(
        alpha=args.alpha,
        bt=args.bootstrap,
        n=args.iterations,
        lbd=args.lbd,
        method=args.method,
    )

    g = s.get_graph_pred()
    np.savetxt(f"{args.output}", g, delimiter=",", fmt="%.3f")
    print(f"Graph prediction was saved to {args.output}.")

    if args.plot:
        s.plot_graph()
        plt.show()


def main(parser):
    """
    Calls the desired function.

    Arguments:
        parser (argparse.ArgumentParser): holding information of the CLI input
    """
    args = parser.parse_args()
    if args.create_data:
        create_data(args)
    elif args.create_graph:
        create_graph(args)
    elif args.snha:
        snha(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="St. Nicolas House Alogrithm CLI",
        description=__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("--snha", action="store_true")
    parser.add_argument("-d", "--data", type=str)
    parser.add_argument("--alpha", type=float, default=0.1)
    parser.add_argument("--bootstrap", action="store_true")
    parser.add_argument("--lbd", type=float, default=0.5)
    parser.add_argument(
        "--method",
        type=str,
        choices=["pearson", "kendall", "spearman"],
        default="pearson",
    )

    parser.add_argument("--create-data", action="store_true")
    parser.add_argument("-g", "--graph")
    parser.add_argument("-s", "--steps", default=25, type=int)
    parser.add_argument("-i", "--iterations", default=200, type=int)
    parser.add_argument("--mean", default=100, type=float)
    parser.add_argument("--std", default=2, type=float)
    parser.add_argument("--noise", default=1, type=float)
    parser.add_argument("--prop", default=0.05, type=float)

    parser.add_argument("--create-graph", action="store_true")
    parser.add_argument(
        "-t",
        "--graph-type",
        type=str,
        choices=[
            "werner",
            "rnd_chain",
            "band",
            "circle",
            "hub",
            "random",
            "barabasi_m1",
            "barabasi_m2",
        ],
        default="werner",
    )
    parser.add_argument("-n", "--nodes", type=int, default=5)
    parser.add_argument(
        "-m", "--mode", type=str, choices=["directed", "undirected"], default="directed"
    )
    parser.add_argument("-e", "--edges", type=int, default=10)
    parser.add_argument("--cont", type=int, default=2)

    parser.add_argument("-p", "--plot", action="store_true")
    parser.add_argument("-o", "--output", type=str)

    main(parser)
